from glob import glob
from multiprocessing import Pool, cpu_count
import os
from typing import Dict
from tqdm.auto import tqdm
import numpy as np
import pandas as pd
from tabulate import tabulate
import csv
import gzip


def compute_header(statistics: Dict[str, bool]) -> str:
    """Return the header for extracted epigenomes.

    Parameters
    ------------------
    statistics:Dict[str, bool]:
        The statistics to be computed

    Returns
    ------------------
    The header separated with tabs.
    """
    return "\t".join([
        "chrom", "chromStart", "chromEnd", "strand",
        *[s for s, enabled in statistics.items() if enabled]
    ])+'\n'


def get_callback(statistic: str):
    return {
        "mean": np.mean,
        "var": np.mean,
        "max": np.max,
        "min": np.min,
        "median": np.median
    }[statistic]

def get_target_path(source:str):
    root, filename = os.path.split(source)
    return "{root}/parsed/{filename}".format(
        root=root,
        filename=filename
    )

def parse_extracted_epigenome(source: str, statistics: Dict[str, bool]):
    """Parse the given source bed-like file."""
    target = get_target_path(source)

    header = compute_header(statistics)
    callbacks = [
        get_callback(s)
        for s, enabled in statistics.items()
        if enabled
    ]
    nans = ["nan"]*len(callbacks)
    os.makedirs(os.path.dirname(target), exist_ok=True)
    with gzip.open(source, "rt") as s:
        with gzip.open(target, "wt") as t:
            # Opening file reader
            reader = csv.reader(s, delimiter='\t')
            # Starting by writing the head
            t.write(header)
            # And now we parse the lines one by one
            for row in reader:
                # We extract the values
                chrom, chromStart, chromEnd, _, _, strand = row[:6]
                # Convert the scores to float values
                scores = np.array([
                    float(s)
                    for s in row[7:]
                    if s != "NA"
                ])
                if scores.size != 0:
                    metrics = [
                        cal(scores).astype(str)
                        for cal in callbacks
                    ]
                else:
                    metrics = nans
                # And write the results
                t.write("\t".join([
                    chrom, chromStart, chromEnd, strand,
                    *metrics
                ])+'\n')


def _parse_extracted_epigenome(kwargs: Dict):
    return parse_extracted_epigenome(**kwargs)


def mine(root: str, statistics: Dict[str, bool]):
    tasks = [
        {
            "source": source,
            "statistics": statistics
        }
        for source in glob(f"{root}/*.bed.gz")
        if not os.path.exists(get_target_path(source))
    ]

    with Pool(cpu_count()) as p:
        list(tqdm(
            p.imap(
                _parse_extracted_epigenome,
                tasks
            ),
            desc="Parse extracted epigenomes",
            total=len(tasks)
        ))
        p.close()
        p.join()



import pandas as pd
import os
from tqdm.auto import tqdm
from typing import List
from multiprocessing import Pool, cpu_count


def to_dict(accession: str, target:str, root: str):
    df = pd.read_csv(
        "{root}/parsed/{accession}.bed.gz".format(
            root=root,
            accession=accession
        ), sep="\t", index_col=[0, 1, 2, 3]).round(2)
    df.columns = pd.MultiIndex.from_product([[target], df.columns])
    return df


def _to_dict(kwargs):
    return to_dict(**kwargs)


def concatenate(root: str, cell_lines: List[str], workers: int):
    data_table = pd.read_csv(
        "{}/epigenomes.csv".format(os.path.dirname(os.path.abspath(__file__))))
    if workers == -1:
        workers = cpu_count()
    filtered = data_table[data_table.cell_line.isin(cell_lines)]
    for cell_line, group in tqdm(filtered.groupby("cell_line"), leave=False, desc="Concatenating cell lines"):
        path = f"{root}/{cell_line}.csv.gz"
        if os.path.exists(path):
            continue
        tasks = [
            {
                "accession": row.accession,
                "target": row.target,
                "root": root
            }
            for _, row in group.iterrows()
        ]
        with Pool(min(cpu_count(), workers)) as p:
            concatenation = pd.concat(list(tqdm(
                p.imap(
                    _to_dict,
                    tasks
                ),
                desc="Concatenating files",
                total=len(tasks),
                leave=False
            )), axis=1)
            p.close()
            p.join()

        concatenation.reset_index().to_csv(path, index=False)



from .extract import extract
from .mine import mine
from .concatenate import concatenate
from typing import List

def build(
    bed_path:str,
    cell_lines:List[str],
    epigenomes_path:str="epigenomes",
    targets_path:str="targets",
    clear_download:bool=False,
    extraction_workers=-1,
    concatenation_workers=-1,
    mine_max:bool=True,
    mine_min:bool=False,
    mine_mean:bool=False,
    mine_median:bool=False,
    mine_variance:bool=False
):
    extract(
        bed_path,
        cell_lines,
        epigenomes_path=epigenomes_path,
        targets_path=targets_path,
        clear_download=clear_download,
        workers=extraction_workers
    )
    mine(targets_path, {
        "max":mine_max,
        "min":mine_min,
        "mean":mine_mean,
        "median":mine_median,
        "variance":mine_variance,
    })
    concatenate(
        targets_path,
        cell_lines,
        concatenation_workers
    )